import praw
from collections import Counter
from time import time, sleep
import urllib2
import nltk
from nltk.stem.porter import PorterStemmer
from string import punctuation
from praw.handlers import MultiprocessHandler


class RedditWordCounter(object):

    def __init__(self, user, stopwords='words/stopwords_english.txt', multiprocess=False):
        handler = MultiprocessHandler() if multiprocess else None
        self.user_agent = 'redditvocab/0.1 bot by {0}'.format(user)
        self.reddit = praw.Reddit(user_agent=self.user_agent, handler=handler)
        self.stemmer = PorterStemmer()

        # Load stop-words
        with open(stopwords, 'rb') as stopwords_file:
            self.stopwords = [word.strip('\n') for word in stopwords_file.readlines()]

    def subreddit_comments(self, subreddit_name, limit=1000, stemming=False, get_all_comments=False):

        def get_vocabulary(comments):

            vocab = Counter()
            num_comments = 0
            for comment in comments:
                if isinstance(comment, praw.objects.Comment):
                    try:
                        # Get the word counts for the comment
                        vocab += self.get_word_count(comment.body, stemming=stemming)
                        num_comments += 1

                    except ValueError:
                        pass
                elif isinstance(comment, praw.objects.MoreComments) and get_all_comments:
                    new_vocab, num_new_comments = get_vocabulary(comment.comments)
                    vocab += new_vocab
                    num_comments += num_new_comments

            return vocab, num_comments

        subreddit = self.reddit.get_subreddit(subreddit_name)
        print "Comments processed for subreddit '{}': 0".format(subreddit_name),

        # Initialise loop variables
        vocabulary = Counter()
        comments_processed = 0

        for submission in subreddit.get_hot(limit=None):
            comments = praw.helpers.flatten_tree(submission.comments)

            # Run over all comments
            submission_vocab, num_new_comments = get_vocabulary(comments)
            vocabulary += submission_vocab
            comments_processed += num_new_comments

            print "{}...".format(comments_processed),

            if limit and comments_processed >= limit:
                break

        print "{}. Finished!".format(comments_processed)
        return vocabulary

    def subreddit_titles(self, subreddit_name, limit=1000, stemming=False):

        subreddit = self.reddit.get_subreddit(subreddit_name)
        print "Titles processed for subreddit {}: 0".format(subreddit_name),

        # Initialise loop variables
        vocabulary = Counter()
        submissions_processed = 0

        for submission in subreddit.get_hot(limit=limit):
            try:
                # Update the word counter to include the comment
                vocabulary += self.get_word_count(submission.title, stemming=stemming)
                submissions_processed += 1

                if submissions_processed % 100 == 0:
                    print "{}...".format(submissions_processed),

            except ValueError:
                pass

        print "{}. Finished!".format(submissions_processed)
        return vocabulary

    def user_comments(self, username, limit=1000, stemming=False):
        user = self.reddit.get_redditor(username)

        vocabulary = Counter()
        comments_processed = 0
        for comment in user.get_comments(limit=limit):
            try:
                # Get the word counts for the comment
                vocabulary += self.get_word_count(comment.body, stemming=stemming)
                comments_processed += 1

                if comments_processed % 100 == 0:
                    print "{0} comments processed for user " \
                          "'{1}'...".format(comments_processed, username)

            except ValueError:
                pass

        print "Processed {0} comments for user {1}".format(comments_processed, username)
        return vocabulary

    def get_word_count(self, text, stop_words=True, stemming=False):
        text = text.lower()
        punctuation_removed = self.remove_punctuation(text)
        tokens = nltk.word_tokenize(punctuation_removed)

        # Remove stop words
        if stop_words:
            tokens = self.remove_stopwords(tokens)

        if stemming:
            tokens = self.stem_tokens(tokens)

        return Counter(tokens)

    @staticmethod
    def remove_punctuation(text, replacement=' ', exclude=""):
        """Remove punctuation from an input string """
        for p in set(list(punctuation)) - set(list(exclude)):
            text = text.replace(p, replacement)

        text = ' '.join(text.split())  # Remove excess whitespace
        return text

    def remove_stopwords(self, tokens):
        return [word for word in tokens if word not in self.stopwords]

    def stem_tokens(self, tokens):
        return [self.stemmer.stem(word) for word in tokens]

    def check_connection(self, timeout=10):
        """Wait for a server response."""
        header = {'User-Agent': self.user_agent}
        start = time()
        while True:
            try:
                request = urllib2.Request("http://www.reddit.com/", headers=header)
                response = urllib2.urlopen(request)
                response.read()
                sleep(2)  # Adhere to Reddit API rule of 30 requests per minute
                if response.getcode() == 200:
                    return True
            except urllib2.HTTPError as err:
                print err
            finally:
                if time() - start > timeout:
                    return False


import config
from redditcounter import RedditWordCounter
from tfidf import TfidfCorpus
import requests
import os
from collections import deque

if __name__ == '__main__':

    def get_subreddit_vocabularies():
        # Initialise Reddit word counter instance
        reddit_counter = RedditWordCounter(config.USERNAME)

        # Initialise tf-idf corpus instance
        corpus_path = os.path.join(config.SAVE_DIR, 'corpus.json')
        comment_corpus = TfidfCorpus(corpus_path)

        # Extract the vocabulary for each of the subreddits specified
        subreddit_queue = deque([subreddit for subreddit in config.SUBREDDITS])
        while len(subreddit_queue) > 0:
            subreddit = subreddit_queue.popleft()

            try:
                vocabulary = reddit_counter.subreddit_comments(subreddit, limit=config.COMMENTS_PER_SUBREDDIT)
            except requests.exceptions.HTTPError as err:
                print err
                # Add subreddit back into queue
                subreddit_queue.append(subreddit)
                continue

            comment_corpus.add_document(vocabulary, subreddit)
            comment_corpus.save_corpus()

        return comment_corpus, corpus_path

    def save_subreddit_top_terms(corpus):
        # Save the top terms for each subreddit in a text file
        save_path = os.path.join(config.SAVE_DIR, 'top_words.txt')
        for document in corpus.get_document_list():
            top_terms = corpus.get_top_terms(document)
            top_terms = sorted(top_terms.items(), key=lambda x: x[1], reverse=True)
            with open(save_path, 'ab') as f:
                f.write(document.encode('utf-8') + '\n' +
                        '\n'.join(['{0}, {1}'.format(term.encode('utf-8'), weight) for term, weight in top_terms])
                        + '\n\n')

        return save_path

    def get_swearword_counts(corpus):
        with open('words/swearwords_english.txt', 'rb') as f:
            swearwords = [word.strip('\n') for word in f.readlines()]

        swearword_counts = dict()
        for document in corpus.get_document_list():
            swearword_counts[document] = corpus.count_words_from_list(document, swearwords)
        return swearword_counts

    def get_vocabulary_sophistication(corpus):
        mean_word_lengths = dict()
        for document in corpus.get_document_list():
            mean_word_lengths[document] = corpus.get_mean_word_length(document)
        return mean_word_lengths

    comment_corpus, corpus_path = get_subreddit_vocabularies()
    print 'TF-IDF corpus saved to %s' % corpus_path

    top_terms_path = save_subreddit_top_terms(comment_corpus)
    print 'Top terms saved to %s' % corpus_path

    swearword_frequency = get_swearword_counts(comment_corpus)
    print 'Normalized swearword frequency:'
    for subreddit, frequency in swearword_frequency.items():
        print '%s, %s' % (subreddit, frequency)

    print 'Average word length by subreddit:'
    word_lengths = get_vocabulary_sophistication(comment_corpus)
    for subreddit, frequency in word_lengths.items():
        print '%s, %s' % (subreddit, frequency)


from __future__ import division
from collections import Counter
import json
import os
import math
import errno
import operator
import numpy as np


class TfidfCorpus(object):
    """Managing feature datasets in an inverted index. Useful for NLP and machine learning applications.
    Corpus takes form:

    To initialise a new TfidfCorpus instance:
    >>> corpus = TfidfCorpus()

    By default the corpus will save to 'tfidf_corpus/corpus.json'. You can specify an existing file to load
    or a specific save path as follows:
    >>> corpus = TfidfCorpus(corpus_path='path/to/corpus.json')

    Data Attributes:
        corpus_path (str): save/load path of the corpus
        document_list (list): list of strings indicating the documents stored in the corpus
        document_lengths (dict): sum of word frequencies contained in each document, takes the form:
            {
                "document1": int,
                "document2": int,
                ...
            }
        corpus (dict): dict of Counters that takes the form:
            {
                "term1": {
                    "document1": int,
                    "document2": int
                },
                "term2": {
                    "document1": int,
                    "document2": int,
                },
                ...
            }

    Methods:
        load_corpus
        get_corpus_path
        get_document_list
        add_document
        get_document
        delete_document
        append_document
        get_idf
        get_tfidf
        get_document_tfidfs
        get_top_terms
        build_feature_array
    """

    def __init__(self, corpus_path='tfidf_corpus/corpus.json'):

        if not corpus_path.lower().endswith('.json'):
            raise Exception('corpus_path provided is not a valid JSON file.')
        make_path(corpus_path)

        self.corpus_path = corpus_path
        self.document_list = list()
        self.document_lengths = dict()
        self.corpus = dict()

        if os.path.isfile(corpus_path):
            self.load_corpus()

    def save_corpus(self):
        """Save the corpus to a JSON file at the path specified in self.corpus_path."""
        with open(self.corpus_path, 'wb') as save_file:
            json.dump(
                {
                    'document_list': self.document_list,
                    'document_lengths': self.document_lengths,
                    'corpus': self.corpus
                },
                save_file
            )

    def load_corpus(self):
        """Load the corpus from a JSON file. File path defined in self.corpus_path."""
        with open(self.corpus_path, 'rb') as load_file:
            data = json.load(load_file)

        try:
            self.document_list = data['document_list']
            self.document_lengths = data['document_lengths']
            self.corpus = data['corpus']

            # Make sure that frequency dicts in corpus are Counter objects
            for term in self.corpus.iterkeys():
                self.corpus[term] = Counter(self.corpus[term])
        except KeyError as err:
            print 'Provided file does not have expected structure'
            raise err

    def get_corpus_path(self):
        return self.corpus_path

    def set_corpus_path(self, path):
        if not path.lower().endswith('.json'):
            raise Exception('Corpus path must be a JSON file (.json extension).')
        self.corpus_path = path

    def get_document_list(self):
        return self.document_list

    def get_vocabulary(self):
        """Return the full list of terms in the corpus."""
        return self.corpus.keys()

    def get_document(self, document_name):
        """Retrieve a document from the corpus."""
        if document_name not in self.document_list:
            raise Exception("No document with name '{0}' found in corpus".format(document_name))
        return Counter({
            term: freqs[document_name] for term, freqs in self.corpus.iteritems() if freqs.get(document_name, 0)
        })

    def add_document(self, document, document_name):
        """Load a document into the corpus.

        :param document: takes the form {'term1': freq1, 'term2', freq2, ...}
        :param document_name: string which uniquely identifies the document
        """
        if document_name in self.document_list:
            print "Document with name '{0}' already exists in corpus." \
                  "Do you wish to replace it?".format(document_name)
            while True:
                replace_doc = raw_input("Response (y/n): ")
                if replace_doc in ['y', 'yes', 'ye']:
                    self.delete_document(document_name)
                    break
                elif replace_doc in ['n', 'no']:
                    return
                else:
                    print 'Could not interpret response. Try again.'

        for term, freq in document.iteritems():
            if not self.corpus.get(term, False):
                self.corpus[term] = Counter()

            self.corpus[term][document_name] = freq

        self.document_list.append(document_name)
        self.document_lengths[document_name] = sum(document.itervalues())

    def delete_document(self, document_name):
        """Delete a document from the corpus.

        :param document_name: string indicating document's name in the corpus - should exist in self.document_list
        """
        if document_name not in self.document_list:
            return
        [freqs.pop(document_name) for term, freqs in self.corpus.iteritems() if freqs.get(document_name, 0)]
        self.document_list.remove(document_name)
        self.document_lengths.pop(document_name)

    def append_document(self, document, document_name):
        """Add new counts to an existing document. If the document doesn't exist in the corpus then it is added.

        :param document: dict or Counter of word counts, e.g. {'i': 1, 'like': 2, 'cheese': 1}
        :param document_name: string indicating document's name in the corpus - should exist in self.document_list
        """
        if document_name not in self.document_list:
            self.add_document(document, document_name)
        else:
            for term, freq in document.iteritems():
                if not self.corpus.get(term, False):
                    self.corpus[term] = Counter()

                self.corpus[term][document_name] += freq

        self.document_lengths[document_name] += sum(document.itervalues())

    def get_idf(self, term):
        """Get inverse document frequency of a given term in the corpus."""
        num_documents = len(self.document_list)
        docs_containing_term = len(self.corpus[term])
        return math.log(num_documents / (1 + docs_containing_term))

    def get_tfidf(self, term, document_name):
        """Get tf-idf score given a term and document in the corpus."""
        tf = self.corpus[term].get(document_name, '') / self.document_lengths[document_name]
        idf = self.get_idf(term)
        return tf * idf

    def get_document_tfidfs(self, document_name, l2_norm=True):
        """Get tf-idf scores for all terms in a document.

        :param document_name: string indicating document's name in the corpus - should exist in self.document_list
        :param l2_norm: if True, applies Euclidean normalization to tf-idf scores of the document
        :return: Counter of tf-idf scores for each term
        """
        tfidfs = {
            term: self.get_tfidf(term, document_name) for term, freq in self.corpus.iteritems()
            if freq.get(document_name, '')
        }

        if l2_norm:
            normalization = np.linalg.norm(tfidfs.values(), axis=0)
            for key, value in tfidfs.items():
                tfidfs[key] = value / normalization

        return Counter(tfidfs)

    def get_top_terms(self, document_name, num_terms=30):
        """Get the top terms for a given document by tf-idf score.

        :param document_name: string indicating document's name in the corpus - should exist in self.document_list
        :param num_terms: number of top terms to return (30 by default)
        :return: dict of top terms and their corresponding tf-idf scores
        """
        tfidfs = self.get_document_tfidfs(document_name)
        sorted_tfidfs = sorted(tfidfs.items(), key=operator.itemgetter(1), reverse=True)
        return dict(sorted_tfidfs[:num_terms])

    def build_feature_array(self, tfidf=True):
        from sklearn.feature_extraction import DictVectorizer
        from sklearn.feature_extraction.text import TfidfTransformer

        index = [self.get_document(document) for document in self.document_list]
        vectorizer = DictVectorizer()
        vectorizer.fit_transform(index)

        if tfidf:
            vectorizer = TfidfTransformer().fit_transform(vectorizer)

        return vectorizer.toarray()

    def count_words_from_list(self, document_name, word_list, normalize=True):
        document = self.get_document(document_name)
        word_counts = [document[word] for word in word_list]
        total_count = sum(word_counts)
        if normalize:
            total_count /= self.document_lengths[document_name]

        return total_count

    def get_mean_word_length(self, document_name, upper_limit=12):
        document = self.get_document(document_name)
        return sum([len(term) * freq for term, freq in document.iteritems()
                    if len(term) <= upper_limit]) / sum(document.itervalues())


def make_path(path):
    """Check if path exists. If it doesn't, create the necessary folders."""

    # Remove file name from path
    base_name = os.path.basename(path)
    if '.' in base_name:
        path = path[:-len(base_name)]

    if not os.path.exists(path):
        try:
            os.makedirs(path)
        except OSError as exception:
            if exception.errno != errno.EEXIST:
                raise

    return path


import os
from setuptools import setup

PACKAGE_NAME = 'redditnlp'
VERSION = '0.1'


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    author='Jai Juneja',
    author_email='jai.juneja@gmail.com',
    description=('A tool to perform natural language processing of reddit content.'),
    license = "BSD",
    keywords = "reddit language processing machine learning",
    url = "https://github.com/jaijuneja/reddit-nlp",
    packages=[PACKAGE_NAME],
    long_description=read('README.md'),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python'
    ],
    install_requires=[
        'praw==2.1.19',
        'nltk==3.0.0',
        'numpy>=1.8.0',
        'scikit-learn==0.15.2'
    ],
    package_data={PACKAGE_NAME: ['words/*.txt']},
    test_suite='tests'
)